#include <errno.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "proc.h"

typedef struct pr {
	struct pr *prev,*next;
	int pid;
	int status;
	int fds[6];
	int socks;

	int ecode; /* exit reason (exit/signal) */
	int earg;  /* exit code or signal name */
} pr_t;

static pr_t head= { &head, &head };

static int sig_installed=0;
static struct sigaction  old_sa;

#ifdef proc_self_test

static char *code_to_str(int code) ;
#endif

	
/*
	one-shot execution: 
 	execute the program with the given input string, collect output and error streams and return
	monitored executon:
		start the program
		enable the user to read/write the output/error/input streams
		enable the user to kill/stop the program
		let the user know whether the program is still alive or not
*/

int proc_kill(int id,int sig) {
	if (kill(id,sig))
		return -errno;
	return 0;
}



static void close_force(int fd) {
	if (fd!=-1)
		while(close(fd) && errno==EINTR)
			;
}


static void close_sockets(pr_t *p) {
	int i;
	for(i=0;i<6;i++) {
		close_force(p->fds[i]);
		p->fds[i]= -1;
	}
}


/* FIXME: in case of errors, put a descriptive text into the error stream
          mini-shell is a good name for this purpose
*/


static void child(pr_t *p,char *fn,char **argv,int opts) {
	int ret= 127;

	

	if (opts & PROC_INP) {
		while (dup2(p->fds[0],0)<0)
			if (errno!=EINTR)
				goto duperr;
		close_force(p->fds[1]);
	}
	ret--;

	if (opts & PROC_OUTP) {
		while (dup2(p->fds[3],1)<0)
			if (errno!=EINTR)
				goto duperr;
		close_force(p->fds[2]);
	}
	ret--;

	if (opts & PROC_ERR)  {
		while (dup2(p->fds[5],2)<0)
			if (errno!=EINTR)
				goto duperr;
		close_force(p->fds[4]);
	}
	ret--;

	execvp(fn,argv);
duperr:
	exit(ret);
}

static pr_t *find(int pid) {
	pr_t *p;
	for(p=head.next;p!=&head;p=p->next)
		if (p->pid==pid)
			return p;
	return NULL;
}


static void child_sig(int sno,struct siginfo *si,void *dummy) {
	int pid= si->si_pid;
	int status= si->si_status;
	int code= si->si_code;
	pr_t *p;
	p= find(pid);
	if (!p)
		return ;
	switch(code) {
		case CLD_EXITED:
		case CLD_KILLED:
		case CLD_DUMPED:
			p->status= PROC_DEAD;
			p->ecode= code;
			p->earg= status;
#ifdef proc_self_test
			printf("child exit: code= %d (%s) arg= %d\n",code,code_to_str(code),status);
			fflush(0);
#endif
			break;
		case CLD_TRAPPED:
		case CLD_STOPPED:
			p->status= PROC_STOPPED;
			break;
		case CLD_CONTINUED:
			p->status= PROC_RUNNING;
			break;
	}
}

static void install_handler() {
	struct sigaction sa;
	if (sig_installed)
		return ;
	sa.sa_sigaction= child_sig;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags= SA_SIGINFO ;
	sa.sa_restorer= 0;
	while(sigaction(SIGCHLD, &sa, &old_sa) && errno==EINTR)
		;
	sig_installed= 1;
}


static void remove_node(pr_t *p) {
	p->prev->next= p->next;
	p->next->prev= p->prev;
	free(p);
}


static void close_socks_main(pr_t *p) {
	close_force(p->fds[0]);
	close_force(p->fds[3]);
	close_force(p->fds[5]);
	p->fds[0]= -1;
	p->fds[3]= -1;
	p->fds[5]= -1;
}

int proc_start(char *fn,char **argv,int opts) {
	
	int eno;
	pr_t *p;

	int i;
	
	install_handler();
	
	p= malloc(sizeof(*p));
	if (!p)
		return -ENOMEM;
	p->prev= head.prev;
	p->next= &head;
	head.prev->next= p;
	head.prev= p;
	p->status= PROC_RUNNING;

	for(i=0;i<6;i++)
		p->fds[i]= -1;

	if (opts & PROC_INP) {
		if (pipe(p->fds+0))
			goto sockerr;
	}

	if (opts & PROC_OUTP) {
		if (pipe(p->fds+2))
			goto sockerr;
	}

	if (opts & PROC_ERR) {
		if (pipe(p->fds+4))
			goto sockerr;
	}

	p->pid= fork();
	if (p->pid<0)
		goto sockerr;
	if (p->pid==0)
		child(p,fn,argv,opts);
	if (0) {
		int i;
		for(i=0;i<6;i++)
			printf("fd[%d]= %d\n",i,p->fds[i]);
	}
	close_socks_main(p);
	return p->pid;

sockerr:
	eno= errno;
	close_sockets(p);
	remove_node(p);
	return -eno;
}

static void interp_status(int s,pr_t *p) {

	if (WIFEXITED(s)) {
		p->ecode= CLD_EXITED;
		p->earg= WEXITSTATUS(s);
	} else if (WIFSIGNALED(s)) {
		p->ecode= CLD_DUMPED;
		p->earg= WTERMSIG(s);
	}
	p->status= PROC_DEAD;
#ifdef proc_self_test
	printf("child exit (wait): code= %d (%s) arg= %d\n",p->ecode,code_to_str(p->ecode),p->earg);
#endif
}

int proc_end(int id,int *st) {
	int status;
	pr_t *p;
	int c,a;
	p= find(id);
	if (!p)
		return 0;
	close_sockets(p);
	while(waitpid(id,&status,0)<0 && errno==EINTR)
		;
	if (p->status!=PROC_DEAD)
		interp_status(status,p);
	c= p->ecode;
	a= p->earg;
	remove_node(p);
	if (c==CLD_KILLED || c==CLD_DUMPED || (c==CLD_EXITED && a!=0)) {
		if (st)
			*st= a;
		return c;
	}
	return 0;
}




int proc_input(int id) {
	pr_t *p= find(id);
	return p ? p->fds[1] : -ESRCH;
}

int proc_output(int id) {
	pr_t *p= find(id);
	return p ? p->fds[2] : -ESRCH;
}

int proc_error(int id) {
	pr_t *p= find(id);
	return p ? p->fds[4] : -ESRCH;
}

int proc_islive(int id) {
	pr_t *p= find(id);
	return p ? (p->status!=PROC_DEAD) : 0;
}

typedef struct buffer {
	struct buffer *next;
	char data[4088];
	int len;
} buffer_t;


static buffer_t *do_read(int fd, fd_set *set, buffer_t *tail,int *ende) {
	int r;
	*ende= 0;
	if (fd!=-1 && FD_ISSET(fd,set)) {
		int l= sizeof(tail->data)- tail->len;
		while(1) {
			r= read(fd, tail->data+tail->len, l);
			if (r<0 && errno==EINTR)
				continue;
			if (r<=0) {
				*ende= 1;
				return tail; /* FIXME: think of something */
			}
			break;
		}
		tail->len+= r;
		if (tail->len==sizeof(tail->data)) {
			buffer_t *b= malloc(sizeof(*b));
			b->len= 0;
			b->next= 0;
			tail->next= b;
			tail= b;
		}
	}
	return tail;
}

static void collect_buffer(char **d,size_t *l, buffer_t *h) {
	size_t s= 0;
	char *D;
	buffer_t *p;
	if (!d)
		return ;
	for(p=h;p;p=p->next)
		s+= p->len;
	D= malloc(s+1);
	*d= D;
	for(p=h;p;p=p->next) {
		memcpy(D, p->data, p->len);
		D+= p->len;
	}
	p= h->next;
	while(p) {
		buffer_t *n= p->next;
		free(p);
		p= n;
	}
	*l= s;
}

static void add_set(int *max, fd_set *set,int fd) {
	FD_SET(fd,set);
	if (fd>*max) *max= fd;
}


/* one-shot */
int proc_run(char *fn,char **argv,int *arg,char *inp,size_t inp_l,char **outp,size_t *outp_l,char **err,size_t *err_l) {
	int input=0;
	int inp_fd=-1,outp_fd=-1,err_fd=-1;
	int r;
	int opts=0;
	int ch;
	buffer_t errhead;
	buffer_t outhead;
	buffer_t *errb;
	buffer_t *outb;
	pr_t *p;
	if (inp) opts |= PROC_INP;
	if (outp) { *outp= NULL; opts |= PROC_OUTP; }
	if (err) { *err= NULL; opts |= PROC_ERR; }
	ch= proc_start(fn,argv, opts);
	if (ch<0)
		return ch;
	if (inp) inp_fd= proc_input(ch);
	if (outp) outp_fd= proc_output(ch);
	if (err) err_fd= proc_error(ch);
	p= find(ch);
	if (!p)
		exit(fprintf(stderr,"internal error in proc library\n"));
	errhead.next= 0;
	outhead.next= 0;
	errhead.len= 0;
	outhead.len= 0;
	errb= &errhead;
	outb= &outhead;
	if (inp_fd!=-1)
		input= 1;
	while(p->status!=PROC_DEAD && (inp_fd!=-1 || err_fd!=-1 || outp_fd!=-1)) {
		int e;
		int status;
		fd_set read_fd,write_fd;
		int max= 0;
		//  struct timeval tm;
		FD_ZERO(&write_fd);
		FD_ZERO(&read_fd);
		if (input)
			add_set(&max, &write_fd, inp_fd);
		if (err_fd!=-1)
			add_set(&max, &read_fd, err_fd);
		if (outp_fd!=-1)
			add_set(&max, &read_fd, outp_fd);
		// tm.tv_sec= 3;
		// tm.tv_usec= 0;

		if (waitpid(ch,&status,WNOHANG)==ch) {
			interp_status(status,p);
			break;
		}
		r= select( max+1,
                           (err_fd!=-1 || outp_fd!=-1) ? &read_fd : NULL,
                           input ? &write_fd : NULL, NULL,
                           NULL);

		if (r<0) {
			if (errno!=EINTR)
				exit(fprintf(stderr,"select failed: %s\n",strerror(errno)));
			continue;
		}
		if (r==0) {
			if (p->status==PROC_DEAD)
				break;
			continue;
		}
		outb= do_read(outp_fd, &read_fd, outb,&e);
		if (e) outp_fd= -1;
		errb= do_read(err_fd, &read_fd, errb,&e);
		if (e) err_fd= -1;
		if (input && FD_ISSET(inp_fd, &write_fd)) {
			int r;
			int l= 1024;
			if (l>inp_l)  
				l= inp_l;
			while( (r=write(inp_fd,inp,l))<0 && errno==EINTR )
				;
			if (r<0)
				break;
			inp_l-= r;
			inp+=r;
			if (inp_l==0) {
				input= 0;
				close_force(p->fds[1]);
				p->fds[1]= -1;
				inp_fd= -1;
			}
		}
	}
	
	while(err_fd!=-1 || outp_fd!=-1) { /* read the remaining data */
		int e;
		fd_set read_fd;
		int max= 0;
		struct timeval tm;
		FD_ZERO(&read_fd);
		if (err_fd!=-1)
			add_set(&max, &read_fd, err_fd);
		if (outp_fd!=-1)
			add_set(&max, &read_fd, outp_fd);
		tm.tv_sec= 0;
		tm.tv_usec= 0;

		r= select( max+1,  &read_fd ,  NULL, NULL, &tm);

		if (r<0) {
			if (errno!=EINTR)
				exit(fprintf(stderr,"select failed: %s\n",strerror(errno)));
			continue;
		}
		if (r==0) 
			break;
		outb= do_read(outp_fd, &read_fd, outb,&e);
		if (e) outp_fd= -1;
		errb= do_read(err_fd, &read_fd, errb,&e);
		if (e) err_fd= -1;
	}

	if (outp) 
		collect_buffer(outp,outp_l, &outhead);
	if (errb)
		collect_buffer(err,err_l, &errhead);
	return proc_end(ch,arg);
}


#ifdef proc_self_test

#include <ctype.h>

static char *code_to_str(int code) {
	switch(code) {
		case CLD_EXITED: return "exit";
		case CLD_KILLED: return "killed";
		case CLD_DUMPED: return "dumped";
	}
	return "unknown";
}

static void subprog()  {
	while(1) {
		int a,b;
		scanf("%d%d",&a,&b);
		if (a==0 && b==0)
			break;
		printf("%d+%d=%d\n",a,b,a+b);
		fflush(0);
	}
}



static void transfer(int in,int out) {
	int r;
	char buf[4096];
	int L;
	do {
		L= read(in,buf,sizeof(buf));
		if (L==0)
			return;
		else
			if (L<0 && errno!=EINTR)
				return ;
	} while(L<=0);

	r= 0;
	while(r<L) {
		int l;
		l= write(out,buf+r,L-r);
		if (l<0 && errno==EINTR)
			continue;
		if (l==0)
			break;
		r+= l;
	}
}

static void test(char **argv) {
	int ch;
	char *fka[3];
	int i=0;
	int inp;
	int outp;
	fka[i++]= argv[0];
	fka[i++]= "-s";
	fka[i++]= 0;
	ch= proc_start(fka[0],fka,PROC_INP | PROC_OUTP);
	if (ch<0) 
		exit(fprintf(stderr,"could not start sub process: %s\n",strerror(-ch)));
	inp= proc_input(ch);
	outp= proc_output(ch);
	while(proc_islive(ch)) {
		int r;
		fd_set readfs;
		FD_ZERO(&readfs);
		FD_SET(0,&readfs);
		FD_SET(outp, &readfs);
		r= select(outp+1, &readfs, 0,0,0);
		if (r<0) {
			if (errno!=EINTR)
				exit(fprintf(stderr,"select failed: %s\n",strerror(errno)));
			continue;
		}
		if (FD_ISSET(0,&readfs))
			transfer(0, inp);
		if (FD_ISSET(outp,&readfs))
			transfer(outp,1);
	}
	printf("\n PROCESS ENDS\n");
	proc_end(ch,NULL);
}

static void run_prog(char **argv,char *inp) {
	char *out;
	size_t l;
	int i;
	int r= proc_run(argv[0],argv,NULL, inp, inp ? strlen(inp) : 0, &out,&l, 0,0);
	if (r)
		exit(fprintf(stderr,"can not execute: %d (%s)\n",r,strerror(r)));
	for(i=0;i<l;i++)
		out[i]=toupper(out[i]);
	fwrite(out,1,l,stdout);
	printf("\n");
}

int main(int argc,char **argv) {
	char *input=0;
	if (argc>1) {
		if (!strcmp(argv[1],"-y")) {
			int i;
			input= argv[2];
			for(i=3;i<argc;i++) {
				argv[i-2]= argv[i];
			}
			argc-=2;
			argv[argc]=0;
		}
		if (!strcmp(argv[1],"-s")) {
			subprog();
			return 0;
		}
		if (!strcmp(argv[1],"-x")) {
			run_prog(argv+2,input);
			return 0;
		}
	}
	test(argv);
	return 0;
}

#endif
